/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Robby Dermody <robbyd@u20.org>
 * Original code by Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Application that can stream audio data from a URL (via HTTP) in a variety of formats, uses wget.
 * 
 * \ingroup applications
 */
 
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>

#include <curl/curl.h>
#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 1.02 $")

#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/logger.h"
#include "asterisk/channel.h"
#include "asterisk/frame.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/translate.h"
#include "asterisk/options.h"
#include "asterisk/app.h"

static char *tdesc = "Play HTTP streams";

static char *app = "HTTPPlayback";

static char *synopsis = "Play an audio stream over HTTP of the specifed format";

static char *descrip = 
"  HTTPPlayback(location,audioFormat,[timeout])\n"
"Uses libcurl to get the audio data at the given location, which is a URL.\n"
"User can exit by pressing any key on the dialpad, or by hanging up.\n"
"Return values: 0: No more data to stream; -1: User hung up;\n"
"Otherwise the ASCII code of the DTMF key the user pressed.\n"
"Valid values of \"audioFmt\" are:\n"
"\"slinear\", \"ulaw\", \"alaw\", \"g726\", \"adpcm\", \"gsm\", \"ilbc\",\n"
"\"speex_cbr_q3or4\", \"speex_cbr_q5or6\", \"speex_cbr_q7or8\", \"speex_cbr_q9\",\n"
"\"speex_cbr_q10\", \"lpc10\", \"g729a\", \"g7231\".\n"
"timeout, when specified (in seconds), alters the length of time that that application will wait for data on\n"
"an inactive HTTP stream before giving up and assuming an end of stream situation. If unspecified, this\n"
"value defaults to 2000 (ms).\n"
"NOTE: With speex, \"speex_cbr_q3or4\" for example means Speex in CBR mode with a\n"
"quality code of either 3 or 4\n";

static int DATA_BUF_SIZE = CURL_MAX_WRITE_SIZE * 2;

struct MyCBStruct {
	struct ast_channel *chan;
	char databuf[DATA_BUF_SIZE];
	size_t bytes_used;
	struct ast_frame f;
	int audiofmt;
	int unitsize;
	int numsamplesinunit;
	int app_return_code;
};

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

static size_t WriteFrameCallback(void *ptr, size_t size, size_t nmemb, void *data)
{
	register int realsize = size * nmemb;
	struct MyCBStruct *cbstruct = (struct MyCBStruct *)data;
	struct ast_frame *f;
	int numunitstowrite = 0;
	int lentowrite = 0;
	int lenleftover = 0;

	/* Check if we got any user input on the channel */
	f = ast_read(chan);
	if (!f) {
		ast_log(LOG_DEBUG, "Null frame == hangup() detected\n");
		res = -1;
		break;
	}
	if (f->frametype == AST_FRAME_DTMF) {
		ast_log(LOG_DEBUG, "User pressed a key: %c\n", f->subclass);
		res = f->subclass; /* return the key that the user pressed (as an int) */
		ast_frfree(f);
		break;
	}
	ast_frfree(f);	
	

	if (! realsize)
		return 0;
	
	/* if there are no bytes used in the data buffer,
	 * we can just write directly to the channel */
	if (! cbstruct->bytes_used) {
		numunitstowrite = realsize / cbstruct->unitsize;
		lentowrite = numunitstowrite * cbstruct->unitsize;
		lenleftover = realsize - lentowrite;

		/* write as much as we can to asterisk (of multiples of size cbstruct->unitsize) */
		cbstruct->f.datalen = lentowrite;
		cbstruct->f.data = ptr;
		cbstruct->f.samples = ast_codec_get_samples(&cbstruct->f);
		if (ast_write(cbstruct->chan, &cbstruct->f) < 0) {
			ast_log(LOG_ERROR, "Channel write error\n");
			return -1;
		}

		/* with whatever we have leftover, place that in the databuf */
		memcpy(cbstruct->databuf, ((char*)ptr) + lentowrite, lenleftover);
		cbstruct->bytes_used = lenleftover;
	}
	else {
		/* with bytes in the data buffer, copy what we've received
		 * into that same buffer and read out of that*/
		if (cbstruct->bytes_used + realsize > DATA_BUF_SIZE)	{
			ast_log(LOG_ERROR, "Aborting to avoid buffer overflow\n");
			return -1;
		}
		
		memcpy(((char*)cbstruct->databuf) + cbstruct->bytes_used, ptr, realsize);
		cbstruct->bytes_used += realsize;

		numunitstowrite = cbstruct->bytes_used / cbstruct->unitsize;
		lentowrite = numunitstowrite * cbstruct->unitsize;
		lenleftover = cbstruct->bytes_used - lentowrite;
		
		/* write as much as we can to asterisk (of multiples of size cbstruct->unitsize) */
		cbstruct->f.datalen = lentowrite;
		cbstruct->f.data = cbstruct->databuf;
		cbstruct->f.samples = ast_codec_get_samples(&cbstruct->f);
		if (ast_write(cbstruct->chan, &cbstruct->f) < 0) {
			ast_log(LOG_ERROR, "Channel write error\n");
			return -1;
		}
		
		/* move whatever is leftover to the start of databuf */
		memmove(cbstruct->databuf, ((char*)cbstruct->databuf) + lentowrite, lenleftover);
		cbstruct->bytes_used = lenleftover;
	}
	
	return realsize;
}

static int curl_internal(struct ast_channel *chan, char *url, int audiofmt, int unitsize, int numsamplesinunit, int timeout)
{
	CURL *curl;
	struct MyCBStruct cbstruct;

    /* initialize curl */
	curl_global_init(CURL_GLOBAL_ALL);
	curl = curl_easy_init();

	if (!curl) {
		ast_log(LOG_ERROR, "Could not initialize libcurl\n");
		return -1;
	}

	curl_easy_setopt(curl, CURLOPT_URL, url);
	curl_easy_setopt(curl, CURLOPT_HEADER, 0); /* don't include HTTP header in output */
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, WriteFrameCallback);
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *)&cbstruct);
	curl_easy_setopt(curl, CURLOPT_BUFFERSIZE, unitsize);
	/* we'd LIKE to get unitsize bytes on every call to our callback */
	curl_easy_setopt(curl, CURLOPT_USERAGENT, "asterisk-libcurl-agent/1.0");
	curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, timeout);

	/*initialize cbstruct */
	/*(cbstruct.f.samples and cbstruct.f.datalen will be
	 * set on every call to our BufferDataCallback) */
	cbstruct.chan = chan;
	cbstruct.app_return_code = 0;
	cbstruct.bytes_used = 0;
	cbstruct.f.frametype = AST_FRAME_VOICE;
	cbstruct.f.subclass = audiofmt;
	cbstruct.f.mallocd = 0;
	cbstruct.f.offset = AST_FRIENDLY_OFFSET;
	cbstruct.f.src = __PRETTY_FUNCTION__;
	cbstruct.f.delivery.tv_sec = 0;
	cbstruct.f.delivery.tv_usec = 0;

	curl_easy_perform(curl);
	curl_easy_cleanup(curl);
	
	return cbstruct.;
}

static int httpplay_exec(struct ast_channel *chan, void *data)
{
	int res=0;
	struct localuser *u;
	int owriteformat;

	char *tmp = NULL;
	int option_audiofmt = 0;
	int option_unitsize = 0;
	int option_numsamplesinunit = 0;
	int option_timeout = 2; /* 2 seconds by default */
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(url);
		AST_APP_ARG(audiofmt);
		AST_APP_ARG(timeout);
	);
	
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "HTTP Audio Playback requires an argument (URL)\n");
		return -1;
	}

	LOCAL_USER_ADD(u);

	/* parse args */
	tmp = ast_strdupa(data);
	if (!tmp) {
		ast_log(LOG_ERROR, "Out of memory!\n");
		LOCAL_USER_REMOVE(u);
		return -1;	
	}

	AST_STANDARD_APP_ARGS(args, tmp);

	/* Verify that required arguments are present */
	if (!args.url || strncasecmp((char *)args.url, "http://", 7)) {
		ast_log(LOG_ERROR, "Stream URL must start with \"http://\"\n");
		LOCAL_USER_REMOVE(u);
		return -1;	
	}	
 	if (!args.audiofmt) {
		ast_log(LOG_ERROR, "Audio format parameter must be specified!\n");
		LOCAL_USER_REMOVE(u);
		return -1;	
 	}

	/* Parse out audio format argument */
	if (strcasestr(args.audiofmt, "slinear"))
		option_audiofmt = AST_FORMAT_SLINEAR;
	else if (strcasestr(args.audiofmt, "ulaw"))
		option_audiofmt = AST_FORMAT_ULAW;
	else if (strcasestr(args.audiofmt, "alaw"))
		option_audiofmt = AST_FORMAT_ALAW;
	else if (strcasestr(args.audiofmt, "g726")) 
		option_audiofmt = AST_FORMAT_G726;
	else if (strcasestr(args.audiofmt, "adpcm")) 
		option_audiofmt = AST_FORMAT_ADPCM;
	else if (strcasestr(args.audiofmt, "gsm")) 
		option_audiofmt = AST_FORMAT_GSM;
	else if (strcasestr(args.audiofmt, "ilbc")) 
		option_audiofmt = AST_FORMAT_ILBC;
	else if (strcasestr(args.audiofmt, "speex"))
		option_audiofmt = AST_FORMAT_SPEEX;
	else if (strcasestr(args.audiofmt, "lpc10"))
		option_audiofmt = AST_FORMAT_LPC10;
	else if (strcasestr(args.audiofmt, "g729a"))
		option_audiofmt = AST_FORMAT_G729A;
	else if (strcasestr(args.audiofmt, "g7231"))
		option_audiofmt = AST_FORMAT_G723_1;
	
	/*choose a buffer size that covers 20ms of data in the specified format
	  20ms of data is 160 samples @ 8HKz, since 8KHz/20ms=160
	  For some formats, 20ms of data doesn't lie on byte boundaries, so we will
	  go up to 30 or 40ms
	  See http://www.openh323.org/docs/bandwidth.html
	  and http://www.voip-info.org/wiki-Codecs
	  and http://www.voip-info.org/wiki/index.php?page=Asterisk+config+codecs.conf
	*/
	if (strcasestr(args.audiofmt, "g7231")) {
		/*30ms*/
 	    option_unitsize = 16;
	    option_numsamplesinunit = 240;
	}
	else if (strcasestr(args.audiofmt, "speex_cbr_q3or4")) {
		/* 8kbps, 20ms */
	    option_unitsize = 20;
	    option_numsamplesinunit = 160;
	}
	else if (strcasestr(args.audiofmt, "speex_cbr_q5or6")) {
		/* 11kbps, 40ms */
	    option_unitsize = 55;
	    option_numsamplesinunit = 320;
	}
	else if (strcasestr(args.audiofmt, "speex_cbr_q7or8")) {
		/* 15kbps, 40ms */
	    option_unitsize = 75;
	    option_numsamplesinunit = 320;
	}
	else if (strcasestr(args.audiofmt, "speex_cbr_q9")) {
		/* 18.2kbps, 40ms */
	    option_unitsize = 91;
	    option_numsamplesinunit = 320;
	}
	else if (strcasestr(args.audiofmt, "speex_cbr_q10")) {
		/* 24.6kbps, 40ms */
	    option_unitsize = 123;
	    option_numsamplesinunit = 320;
	}
	else if (strcasestr(args.audiofmt, "lpc10")) {
		/* 22.5ms (weirdo in the group) */
	    option_unitsize = 7;
	    option_numsamplesinunit = 180;
	}
	else {
		/* ast_codec_get_len() covers the rest of the codecs */
	    option_unitsize = ast_codec_get_len(option_audiofmt, 160);
	    option_numsamplesinunit = 160;
	}
	
	/* parse out timeout argument (if specified) */
	if(args.timeout) {
		option_timeout = atoi(args.timeout);
	}

	ast_stopstream(chan);

	owriteformat = chan->writeformat;
	res = ast_set_write_format(chan, option_audiofmt);
	if (res < 0) {
		ast_log(LOG_WARNING, "Unable to set write format to %i\n", option_audiofmt);
		LOCAL_USER_REMOVE(u);
		return -1;
	}
	
	ast_log(LOG_DEBUG, "Playing URL \"%s\" with file format \"%s\" (%i). Timeout is %i ms\n",
		    args.url, args.audiofmt, option_audiofmt, option_timeout);
	
	curl_internal(chan, args.url, option_audiofmt, option_unitsize, option_numsamplesinunit, option_timeout);

	if (!res && owriteformat)
		ast_set_write_format(chan, owriteformat);

	LOCAL_USER_REMOVE(u);
	return res;
}


///GET RID OF
	res = httpplay((char *)args.url, fds[1]);
	/* Wait 1000 ms first */
	next = ast_tvnow();
	next.tv_sec += 1;
	if (res >= 0) {
		pid = res;
		for (;;) {
			ms = ast_tvdiff_ms(next, ast_tvnow());
			if (ms <= 0) {
				res = timed_read(fds[0], myf.frdata, sizeof(myf.frdata), option_timeout);
				if (res > 0) {
					myf.f.frametype = AST_FRAME_VOICE;
					myf.f.subclass = option_audiofmt;
					myf.f.datalen = res;
					myf.f.mallocd = 0;
					myf.f.offset = AST_FRIENDLY_OFFSET;
					myf.f.src = __PRETTY_FUNCTION__;
					myf.f.delivery.tv_sec = 0;
					myf.f.delivery.tv_usec = 0;
					myf.f.data = myf.frdata;
					myf.f.samples = ast_codec_get_samples(&myf.f);
					if (ast_write(chan, &myf.f) < 0) {
						res = -1;
						break;
					}
				} else {
					ast_log(LOG_DEBUG, "No more data to read\n");
					res = 0;
					break;
				}
				next = ast_tvadd(next, ast_samp2tv(myf.f.samples, 8000)); /* pause a second? */
			} else {
				ms = ast_waitfor(chan, ms);
				if (ms < 0) {
					ast_log(LOG_DEBUG, "Hangup detected\n");
					res = -1;
					break;
				}
				if (ms) {
					f = ast_read(chan);
					if (!f) {
						ast_log(LOG_DEBUG, "Null frame == hangup() detected\n");
						res = -1;
						break;
					}
					if (f->frametype == AST_FRAME_DTMF) {
						ast_log(LOG_DEBUG, "User pressed a key: %c\n", f->subclass);
						res = f->subclass; /* return the key that the user pressed (as an int) */
						ast_frfree(f);
						break;
					}
					ast_frfree(f);
				} 
			}
		}
	}



int unload_module(void)
{
	int res;

	res = ast_unregister_application(app);

	STANDARD_HANGUP_LOCALUSERS;
	
	return res;
}

int load_module(void)
{
	return ast_register_application(app, httpplay_exec, synopsis, descrip);
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
